// ASSUMPTIONS:
// pop_size is evenly divisible by num_threads

#include <assert.h>
#include <math.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/processor.h>
#include <sys/procset.h>
#include <sys/time.h>
#include <sys/types.h>

#include "barrier.h"
#include "tsp-pthread.h"

#define MAX_PROCS 16

// variables
int processor_ids[MAX_PROCS] = {29, 28, 25, 24, 21, 20, 17, 16, 13, 9, 8, 5, 0, 12, 4, 1};
/*
int processor_ids[MAX_PROCS] = {0, 1, 4, 5, 8, 9, 12, 13, 16, 17, 20, 21, 24,
25, 28, 29};*/
barrier_t bar; // barrier
char* in_file; // name of input file
int pop_size; // population size
int num_gens; // number of generations
float xover_rate; // crossover rate
float mut_rate; // mutation rate
int num_threads; // number of threads
unsigned int rand_seed; // random seed
int num_cities; // number of cities
city_t* cities; // info for all the cities
pop_mem_t* pop; // the population: an array of population members
int max_pop; // the maximum population, including parents & children

// threaded
void breed(int id) {
  int i; // loop counter
  int dad, mom; // parent indices
  float rand_mut; // randomly calculated amount of mutation

  // these two specify this thread's share of the data parallelism
  // this function is divided by pop_size * xover_rate
  int chunk_size = (max_pop - pop_size) / num_threads;
  int my_range_bottom = id * chunk_size + pop_size; 
  int my_range_top = chunk_size + my_range_bottom;

  // breed xover_rate * pop_size new solutions
  for (i = my_range_bottom; i < my_range_top; i++) {
    // select two parent solutions from current population
    dad = rand() * (pop_size - 1) / RAND_MAX;
    do {
      mom = rand() * pop_size / RAND_MAX;
    } while(mom == dad || mom == pop_size);

    // form child solution via crossover
    crossover(mom, dad, i, FALSE);

    // mutate child
    rand_mut = (float) rand() / RAND_MAX;
    if (rand_mut < mut_rate) { // mutate once
      random_swap(i);
      if (rand_mut < 2 * mut_rate / 3) {
        random_swap(i);
        if (rand_mut < mut_rate / 3) {
          random_swap(i);
        }
      }
    }

    // evaluate child's fitness
    calc_fitness(i);
  }
}

void calc_fitness(int idx) {
  int i; // loop counter

  pop[idx].fitness = 0;
  for (i = 0; i < num_cities; i++) {
    pop[idx].fitness += distance(pop[idx].order[i], 
                                 pop[idx].order[(i + 1) % num_cities]);
  }
}

void crossover(int mom, int dad, int child, int echo) {
  int i, tmp; // loop counter
  int mom_start; // where the mom's half gets copied into the child
  int mom_idx; // position in mom's order
  int dad_idx; // position in dad's order
  int selected[num_cities]; // whether that was set by mom

  for (i = 0; i < num_cities; i++) {
    selected[i] = FALSE;
  }

  mom_start = rand() * (num_cities - 1) / RAND_MAX;

  // get from mom
  for (i = 0; i < num_cities / 2; i++) {
    mom_idx = (mom_start + i) % num_cities;
    pop[child].order[i] = pop[mom].order[mom_idx];
    selected[pop[mom].order[mom_idx]] = TRUE;
  }

  // get from dad
  dad_idx = 0;
  while (i < num_cities) {
    if (!selected[pop[dad].order[dad_idx]]) {
      pop[child].order[i] = pop[dad].order[dad_idx];
      i++;
    }
    dad_idx++;
  }

  if (echo) {
    printf("mom: ");
    print_member(mom);
    printf("dad: ");
    print_member(dad);
    printf("child: ");
    print_member(child);
  }
}

void cull() {
  quick_sort(0, max_pop - 1);
}

void deallocate() {
  // local variable
  int i; // loop counter

  // clean up cities
  for (i = 0; i < num_cities; i++) {
    free(cities[i].name);
  }
  free(cities);

  // clean up population
  for (i = 0; i < max_pop; i++) {
    free(pop[i].order);
  }
  free(pop);
}

double distance(int a, int b) {
  float lat_dif = cities[a].lat - cities[b].lat;
  float lon_dif = cities[a].lon - cities[b].lon;
  return sqrt(pow(lat_dif, 2) + pow(lon_dif, 2));
}

// threaded
void gen_pop(int id, int echo) {
  // local variables
  int i, j, tmp; // loop counter
  int selected[num_cities]; // TRUE if corresponding city already
                            // selected for path
  int local_sense = TRUE;
  int city; // the index of the next available city
  int city_idx; // index of the next city along the path
  float norm_rand; // random value from [0, 1]
  int sel_count; // number of selected cities seen so far

  // these two specify this thread's share of the data parallelism
  // this function is divided by pop_size
  int chunk_size;
  int my_range_bottom; 
  int my_range_top;

  // initialize selected
  for (i = 0; i < num_cities; i++) {
    selected[i] = FALSE;
  }

  chunk_size = pop_size / num_threads;
  my_range_bottom = id * chunk_size;
  my_range_top = my_range_bottom + chunk_size;
  // generate initial values for the first generation (no children)
  for (i = my_range_bottom; i < pop_size; i++) {
    j = 0;
    while (j < num_cities) {
      // allow 1 > than desired b/c that seems to show up very rarely:
      // if use actual desired max, is always last in path!
      city = rand() * num_cities / RAND_MAX;
      if (city != num_cities && selected[city] != local_sense) { 
        // not yet selected
        selected[city] = local_sense;
        pop[i].order[j] = city;
        j++;
      }
    }
    local_sense = !local_sense;

    // calculate initial fitness
    calc_fitness(i);
  }

  if (id == 0 && echo) {
    print_pop(FALSE);
  }
}

void parse_args(int argc, char** argv, int echo) {
  // local variables
  FILE* fd; // file descriptor for input file
  int i, j; // loop counters
  char name_buf[MAX_NAME_LENGTH];  // name read from file

  // parse arguments
  if (argc != 8) {
    printf("%s [input file] [population size] ", argv[0]);
    printf("[number of generations]\n[crossover rate] [mutation rate] ");
    printf("[number of threads] [random seed]\n");
    exit(-1);
  }

  in_file = argv[1];
  pop_size = atoi(argv[2]);
  num_gens = atoi(argv[3]);
  xover_rate = atof(argv[4]);
  mut_rate = atof(argv[5]);
  num_threads = atoi(argv[6]);
  rand_seed = atoi(argv[7]);
  if (echo) {
    print_args();
  }

  srand(rand_seed);

  // parse the input file
  fd = fopen(in_file, "r");
  fscanf(fd, "%d", &num_cities);
  cities = (city_t*) malloc(num_cities * sizeof(city_t));
  for (i = 0; i < num_cities; i++) {
    fscanf(fd, "%f %f", &cities[i].lat, &cities[i].lon);
    j = 0;
    fgetc(fd);
    while (j < MAX_NAME_LENGTH) {
      name_buf[j] = fgetc(fd);
      if (name_buf[j] == '\n') {
        name_buf[j] = '\0';
        break;
      }
      j++;
    }
    cities[i].name = (char*) malloc(j * sizeof(char));
    strcpy(cities[i].name, name_buf);
  }
  fclose(fd);
  if (echo) {
    print_cities();
  }
}

void print_args() {
  printf("in_file=\"%s\", ", in_file);
  printf("pop_size=%d, num_gens=%d, ", pop_size, num_gens);
  printf("xover_rate=%f, mut_rate=%f, ", xover_rate, mut_rate);
  printf("num_threads=%d, rand_seed=%d\n", num_threads, rand_seed);
}

void print_cities() {
  int i; // loop counter
  printf("num_cities=%d\n", num_cities);
  for (i = 0; i < num_cities; i++) {
    printf("%d: ", i);
    print_city(i);
  }
}

void print_city(int idx) {
  printf("lat=%f, lon=%f, ", cities[idx].lat, cities[idx].lon);
  printf("name=%s\n", cities[idx].name);
}

void print_member(int idx) {
  int i; // loop counter

  printf("[");
  for (i = 0; i < num_cities - 1; i++) {
    printf("%d, ", pop[idx].order[i]);
  }
  printf("%d] %f\n", pop[idx].order[i], pop[idx].fitness);
}

void print_pop(int all) {
  int i; // loop counter
  int lim; // how many to print

  lim = all ? max_pop : pop_size;

  for (i = 0; i < lim; i++) {
    print_member(i);
  }
}

void print_solution() {
  printf("*** SOLUTION ***\n");
  print_member(0);
}

void print_solution_long() {
  int i; // loop counter

  printf("*** SOLUTION ***\n");
  for (i = 0; i < num_cities; i++) {
    print_city(pop[0].order[i]);
  }
  printf("distance=%f\n", pop[0].fitness);
}

void quick_sort(int left, int right) {
  pop_mem_t pivot;
  int tmpl, tmpr, pivot_idx; // local indices

  tmpl = left;
  tmpr = right;
  pivot = pop[left];
  while (tmpl < tmpr) {
    while (pop[tmpr].fitness >= pivot.fitness && tmpl < tmpr) {
      tmpr--;
    }
    if (tmpl != tmpr) {
      pop[tmpl] = pop[tmpr];
      tmpl++;
    }
    while(pop[tmpl].fitness <= pivot.fitness && tmpl < tmpr) {
      tmpl++;
    }
    if (tmpl != tmpr) {
      pop[tmpr] = pop[tmpl];
      tmpr--;
    }
  }
  pop[tmpl] = pivot;
  pivot_idx = tmpl;
  if (left < pivot_idx) {
    quick_sort(left, pivot_idx - 1);
  }
  if (right > pivot_idx) {
    quick_sort(pivot_idx + 1, right);
  }
}

void random_swap(int pop_idx) {
  int swap1, swap2, tmp; // for swapping cities

  swap1 = rand() * (num_cities - 1) / RAND_MAX;
  swap2 = rand() * (num_cities - 1) / RAND_MAX;
  tmp = pop[pop_idx].order[swap1];
  pop[pop_idx].order[swap1] = pop[pop_idx].order[swap2];
  pop[pop_idx].order[swap2] = tmp;
}

void* thread_main(void* my_id) {
  int cur_gen = 0; // current generation
  int local_sense = FALSE; // for sense-reversing barrier
  int id = (int) my_id; // this thread's unique id [0, n)

  // bind thread to a processor, ensuring at most one thread per proc
  assert(!processor_bind(P_LWPID, P_MYID, processor_ids[id], NULL));
  barrier(&bar, &local_sense);

  gen_pop(id, FALSE);
  /*
  if (id == 0) {
    printf("*** INITIAL POPULATION ***\n");
    print_pop(FALSE);
  }
  */
  barrier(&bar, &local_sense);

  while (cur_gen < num_gens) {
    breed(id);
    barrier(&bar, &local_sense);

    if (id == 0) {
      cull();
      /*
      printf("*** POPULATION GENERATION %d ***\n", cur_gen);
      print_pop(TRUE);
      */
    }
    barrier(&bar, &local_sense);
    cur_gen++;
  }
}

int main(int argc, char** argv) {
  pthread_t* threads;
  pthread_attr_t attr;
  int i; // loop counter
  hrtime_t start_time, end_time;

  parse_args(argc, argv, FALSE);

    // allocate population (enough for children later)
  max_pop = pop_size * (1 + xover_rate);
  pop = (pop_mem_t*) malloc(max_pop * sizeof(pop_mem_t));
  for (i = 0; i < max_pop; i++) {
    pop[i].order = (int*) malloc(num_cities * sizeof(int));
  }

  // initialize thread-related stuff
  threads = (pthread_t*) malloc(num_threads * sizeof(pthread_t));
  pthread_attr_init(&attr);
  pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
  barrier_init(&bar, num_threads);

  start_time = gethrtime();

  // start the threads in the thread_main routine
  for (i = 0; i < num_threads; i++) {
    assert(!pthread_create(&threads[i], &attr, thread_main, (void*) i));
  }

  // as each thread completes, merge with it
  for (i = 0; i < num_threads; i++) {
    assert(!pthread_join(threads[i], NULL));
  }

  end_time = gethrtime();

  printf("%f\n", (end_time - start_time) / 1000000000.0);
  deallocate();

  // clean up thread-related stuff
  pthread_attr_destroy(&attr);
  barrier_destroy(&bar);

  return 0;
}
